#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <gst/gst.h>

int server(const char *fname, int port);
int client(const char *srvhost, int port);

void link_demux_output(GstElement *elem, GstPad *pad, void *data);
int bus_callback(GstBus *bus, GstMessage *msg, void *data);

int main(int argc, char **argv)
{
	int i;
	char *srvhost = 0;
	char *fname = 0;
	int port = 6767;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'p':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-p must be followed by the port number\n");
					return 1;
				}
				port = atoi(argv[i]);
				break;

			case 's':
				fname = argv[++i];
				break;

			case 'c':
				srvhost = argv[++i];
				break;

			default:
				fprintf(stderr, "unrecognized option: %s\n", argv[i]);
				return 1;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return 1;
		}
	}

	if(fname && srvhost) {
		fprintf(stderr, "-c and -s are mutually exclusive\n");
		return 1;
	}
	if(!fname && !srvhost) {
		fprintf(stderr, "must specify either -c or -s\n");
		return 1;
	}

	gst_init(&argc, &argv);

	if(fname) {
		server(fname, port);
	} else {
		client(srvhost, port);
	}
	return 0;

	/*GstBus *bus;
	GstElement *pline, *src, *demux, *decode, *conv, *sink;
	GMainLoop *loop;

	gst_init(&argc, &argv);
	loop = g_main_loop_new(0, 0);

	pline = gst_pipeline_new("foo");
	src = gst_element_factory_make("filesrc", "file-source");
	demux = gst_element_factory_make("oggdemux", "ogg-demuxer");
	decode = gst_element_factory_make("vorbisdec", "vorbis-decoder");
	conv = gst_element_factory_make("audioconvert", "converter");
	sink = gst_element_factory_make("autoaudiosink", "audio-out");
	assert(pline && src && demux && decode && conv && sink);

	g_object_set(G_OBJECT(src), "location", argv[1], NULL);

	bus = gst_pipeline_get_bus(GST_PIPELINE(pline));
	gst_bus_add_watch(bus, bus_callback, loop);
	gst_object_unref(bus);

	gst_bin_add_many(GST_BIN(pline), src, demux, decode, conv, sink, NULL);

	gst_element_link(src, demux);
	gst_element_link_many(decode, conv, sink, NULL);
	g_signal_connect(demux, "pad-added", G_CALLBACK(link_demux_output), decode);

	printf("foo: %s\n", argv[1]);
	gst_element_set_state(pline, GST_STATE_PLAYING);

	g_main_loop_run(loop);

	printf("bar\n");
	gst_element_set_state(pline, GST_STATE_NULL);

	gst_object_unref(GST_OBJECT(pline));
	return 0;
	*/
}


int server(const char *fname, int port)
{
	GstElement *pline, *src, /**demux, */*tcpsink;
	GMainLoop *loop;
	GstBus *bus;

	loop = g_main_loop_new(0, 0);

	pline = gst_pipeline_new("server-pipeline");
	src = gst_element_factory_make("filesrc", "file-source");
	/*demux = gst_element_factory_make("oggdemux", "ogg-demuxer");*/
	tcpsink = gst_element_factory_make("tcpserversink", "tcp-server-sink");

	if(!pline || !src /*|| !demux */|| !tcpsink) {
		fprintf(stderr, "SERVER: failed to create pipeline elements\n");
		return -1;
	}

	g_object_set(G_OBJECT(src), "location", fname, NULL);
	g_object_set(G_OBJECT(tcpsink), "port", port, NULL);
	/*g_object_set(G_OBJECT(tcpsink), "protocol", 1, NULL);*/

	bus = gst_pipeline_get_bus(GST_PIPELINE(pline));
	gst_bus_add_watch(bus, bus_callback, loop);
	gst_object_unref(bus);

	gst_bin_add_many(GST_BIN(pline), src, /*demux, */tcpsink, NULL);

	gst_element_link(src, tcpsink);
	/*gst_element_link(src, demux);
	g_signal_connect(demux, "pad-added", G_CALLBACK(link_demux_output), tcpsink);*/

	printf("SERVER: start streaming: %s\n", fname);
	gst_element_set_state(pline, GST_STATE_PLAYING);

	g_main_loop_run(loop);

	printf("SERVER: stopping\n");
	gst_element_set_state(pline, GST_STATE_NULL);

	gst_object_unref(GST_OBJECT(pline));
	return 0;
}

int client(const char *srvhost, int port)
{
	GstElement *pline, *tcpsrc, *demux, *decode, *conv, *sink;
	GMainLoop *loop;
	GstBus *bus;

	loop = g_main_loop_new(0, 0);

	pline = gst_pipeline_new("client-pipeline");
	tcpsrc = gst_element_factory_make("tcpclientsrc", "tcp-client-source");
	demux = gst_element_factory_make("oggdemux", "ogg-demuxer");
	decode = gst_element_factory_make("vorbisdec", "vorbis-decoder");
	conv = gst_element_factory_make("audioconvert", "converter");
	sink = gst_element_factory_make("autoaudiosink", "audio-sink");

	if(!pline || !tcpsrc || !decode || !conv || !sink) {
		fprintf(stderr, "SERVER: failed to create pipeline elements\n");
		return -1;
	}

	g_object_set(G_OBJECT(tcpsrc), "host", srvhost, NULL);
	g_object_set(G_OBJECT(tcpsrc), "port", port, NULL);
	/*g_object_set(G_OBJECT(tcpsrc), "protocol", 1, NULL);*/

	bus = gst_pipeline_get_bus(GST_PIPELINE(pline));
	gst_bus_add_watch(bus, bus_callback, loop);
	gst_object_unref(bus);

	gst_bin_add_many(GST_BIN(pline), tcpsrc, demux, decode, conv, sink, NULL);
	gst_element_link(tcpsrc, demux);
	g_signal_connect(demux, "pad-added", G_CALLBACK(link_demux_output), decode);
	gst_element_link_many(decode, conv, sink, NULL);

	printf("CLIENT: start playing from: %s:%d\n", srvhost, port);
	gst_element_set_state(pline, GST_STATE_PLAYING);

	g_main_loop_run(loop);

	printf("CLIENT: stopping\n");
	gst_element_set_state(pline, GST_STATE_NULL);

	gst_object_unref(GST_OBJECT(pline));
	return 0;
}

void link_demux_output(GstElement *elem, GstPad *pad, void *data)
{
	GstPad *sinkpad;
	GstElement *decoder = (GstElement*)data;

	printf("linking\n");

	sinkpad = gst_element_get_static_pad(decoder, "sink");
	gst_pad_link(pad, sinkpad);
	gst_object_unref(sinkpad);
}

int bus_callback(GstBus *bus, GstMessage *msg, void *data)
{
	GMainLoop *loop = (GMainLoop*)data;
	GError *err;

	switch(GST_MESSAGE_TYPE(msg)) {
	case GST_MESSAGE_EOS:
		g_main_loop_quit(loop);
		break;

	case GST_MESSAGE_ERROR:
		gst_message_parse_error(msg, &err, 0);
		fprintf(stderr, "error: %s\n", err->message);
		g_error_free(err);
		g_main_loop_quit(loop);
		break;

	default:
		break;
	}

	return 1;
}
